// $Id$

//////////////////////////////////////////////////////////////////////////////
// Copyright (c) Andreas Huber Doenni 2002-2004.
// Use, modification and distribution are subject to the Boost Software
// License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//////////////////////////////////////////////////////////////////////////////

//=============================================================================
/**
 *  @file    CIDRWState.h
 *
 *  @author  Fukasawa Mitsuo
 *
 *
 *    Copyright (C) 2003-2004 BEE Co.,Ltd. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
//=============================================================================


//////////////////////////////////////////////////////////////////////////////
// The following code implements the state-machine:
//
// E99 : CARRIER ID READER/WRITER FUNCTIONAL STANDARD
//
//                                               +------+
//                                               V      |*11
// +-------------------------------------------------+  |
// |      +--------------+      CIDRW                |--+
// |  O-->| INITIALIZING |                           +---------------------------+
// |   *1 +--------------+                           |  *12       HEAD nn        |
// |             |*2          +---------+            | O--->(C)                  |
// |             V            | RUNNING |            |                           |
// | +------------------------+-----+---+----------+ | +-----------------------+ |
// | |    *3         OPERATIONAL    :  *8          | | |O--+                   | |
// | | O----+                       : O-->(C)      | | |   |    OPERATING      | |
// | |      |                       :              | | |   V                   | |
// | | +----|---------------------+ : +----------+ | | | +------+ *13 +------+ | |
// | | |    V        OPERATING    | : |    NO    | | | | |      |---->|      | | |
// | | | +--------+ *4 +--------+ | : |  ALARMS  | | | | | IDLE |     | BUSY | | |
// | | | |  IDLE  |--->|  BUSY  | | : |          | | | | |      |<----|      | | |
// | | | |        |<---|        | | : +----------+ | | | +------+ *14 +------+ | |
// | | | +--------+ *5 +--------+ | :   *9|  ^     | | |                       | |
// | | |    |  ^                  | :     V  |*10  | | +-----------------------+ |
// | | +----|--|------------------+ : +----------+ | |      *15|        ^        |
// | |    *6|  |*7                  : |          | | |         V        |*16     |
// | |      V  |                    : |  ALARMS  | | | +-----------------------+ |
// | | +--------------------------+ : |          | | | |                       | |
// | | |       MAINTENANCE        | : +----------+ | | |     NOT OPERATING     | |
// | | +--------------------------+ : ALARM STATUS | | |                       | |
// | +------------------------------+--------------+ | +-----------------------+ |
// |                                                 |                           |
// +-------------------------------------------------+---------------------------+
//
//  *1: Powerup or reset.
//  *2: Initialize is complete.
//  *3: Default entry into OPERATING.
//  *4: A service request to read or write or perform diagnostics is received.
//  *5: All sevice requests that affect the state of the hardware are completed.
//  *6: A user selects the MAINTENANCE state and all heads are idle.
//  *7: A user selects the OPERATING state and all heads are idle.
//  *8: Default entry into ALARM STATUS.
//  *9: An alarm condition is detected.
// *10: All alarm conditions have cleared.
// *11: A reset service request is received.
// *12: Initialization on powerup or reset.
// *13: Request to perform a sevice.
// *14: Complete of service.
// *15: An alarm condition is detected.
// *16: Self-Diagnostics Service.
//

#ifndef JYUGEM_CIDRW_CIDRWSTATE_H
#define JYUGEM_CIDRW_CIDRWSTATE_H

#ifdef _MSC_VER
#pragma warning( disable: 4267 )
#endif

#include "jyugem/bee/state_machine_task.h"
#include "jyugem/bee/viewed_state.h"
namespace bee = jyugem::bee;


struct EvReset        : fsm::event< EvReset > {};
struct EvInitialized  : fsm::event< EvInitialized > {};
struct EvAlarmSet     : fsm::event< EvAlarmSet > {};
struct EvAlarmClear   : fsm::event< EvAlarmClear > {};
struct EvRead         : fsm::event< EvRead > {};
struct EvWrite        : fsm::event< EvWrite > {};
struct EvDiagnostic   : fsm::event< EvDiagnostic > {};
struct EvComplete     : fsm::event< EvComplete > {};
struct EvMaintenance  : fsm::event< EvMaintenance > {};
struct EvOperating    : fsm::event< EvOperating > {};

//
namespace jyugem { namespace cidrw
{

typedef bee::named_state     NamedState;
typedef bee::viewed_state<> ViewedState;

//
// State Machine
//

class  CIDRW;
class  CIDRWHead;
struct Initialize;

class CIDRWState : public state_machine_task< CIDRWState, Initialize >
{
public:
  typedef bee::state_machine_task< CIDRWState, Initialize >  task_base;

  CIDRWState(CIDRW * cidrw);
  ~CIDRWState() {}

  virtual int send(trigger * trig)
  {
    return send_trigger<ViewedState>(trig->name());
  }

  const cstr_id * get_current_name();

  static void InitEventDictionary();
  static task_base::event_dictionary _event_dictionary;

  size_t m_alarmq;
};

//
// Innor State
//
struct Running;
struct OperationalStatus;
struct AlarmStatus;
struct Idle;
struct Maintenance;
struct Busy;
//struct Operating;
struct InitAlarmStatus;
struct Alarms;
struct NoAlarms;


struct Initialize : fsm::simple_state< Initialize,
  CIDRWState, fsm::transition< EvInitialized, Running > >,
  ViewedState
{
  Initialize() : ViewedState() { TRACE_STATE_ENTER(); }
};

struct Running : fsm::simple_state< Running, CIDRWState,
  fsm::transition< EvReset, Initialize >,
  mpl::list< OperationalStatus, AlarmStatus > >,
  NamedState
{
  Running() : NamedState() { TRACE_STATE_ENTER(); }
};

struct OperationalStatus : fsm::simple_state< OperationalStatus,
  Running::orthogonal<0>, fsm::no_reactions, Idle >,
  NamedState
{
    OperationalStatus() : NamedState() { TRACE_STATE_ENTER(); }

};

// struct Operating : fsm::simple_state< Operating,
//   OperationalStatus, fsm::no_reactions, Idle >,
//   NamedState
// {
//   Operating(): NamedState() { TRACE_STATE_ENTER(); }
// };


struct Idle : fsm::simple_state< Idle, OperationalStatus,
  mpl::list< fsm::transition< EvRead, Busy >,
    fsm::transition< EvWrite, Busy >, fsm::transition< EvDiagnostic, Busy >,
    fsm::transition< EvMaintenance, Maintenance > > >,
  ViewedState
{
    Idle() : ViewedState() { TRACE_STATE_ENTER(); }
};


struct Busy : fsm::simple_state< Busy, OperationalStatus,
  fsm::transition< EvComplete, Idle > >,
  ViewedState
{
    Busy() : ViewedState() { TRACE_STATE_ENTER(); }
};

struct Maintenance : fsm::simple_state< Maintenance,
  OperationalStatus, fsm::transition< EvOperating, Idle > >,
  ViewedState
{
  Maintenance() : ViewedState() { TRACE_STATE_ENTER(); }
};

struct AlarmStatus : fsm::simple_state< AlarmStatus,
  Running::orthogonal<1>, fsm::no_reactions, InitAlarmStatus >,
  NamedState
{
    AlarmStatus() : NamedState() { TRACE_STATE_ENTER(); }
};

struct InitAlarmStatus : fsm::state< InitAlarmStatus, AlarmStatus,
  fsm::custom_reaction< EvEvalCondition > >,
  NamedState
{
  InitAlarmStatus(my_context ctx) : NamedState(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    post_event( boost::intrusive_ptr< EvEvalCondition >( new EvEvalCondition()) );
  }

  fsm::result react( const EvEvalCondition & )
  {
    if (outermost_context().m_alarmq > 0)
      return transit< Alarms >();
    else
      return transit< NoAlarms >();
  }
};

struct Alarms : fsm::simple_state< Alarms, AlarmStatus,
  mpl::list< fsm::custom_reaction< EvAlarmClear >,
    fsm::custom_reaction< EvAlarmSet > > >,
  NamedState
{
  Alarms() : NamedState() { TRACE_STATE_ENTER(); }

  fsm::result react( const EvAlarmClear & );
  fsm::result react( const EvAlarmSet & );
};

struct NoAlarms : fsm::simple_state< NoAlarms, AlarmStatus,
  fsm::custom_reaction< EvAlarmSet > >,
  NamedState
{
  NoAlarms() : NamedState() { TRACE_STATE_ENTER(); }

  fsm::result react( const EvAlarmSet & );
};

//-----------------------------------------------------------------------------
// Header state
//
struct InitHead;
struct HeadOperating;
struct HeadNotOperating;
struct HeadIdle;
struct HeadBusy;

class CIDRWHeadState : public state_machine_task< CIDRWHeadState, InitHead >
{
public:
  typedef bee::state_machine_task< CIDRWHeadState, InitHead >  task_base;

  CIDRWHeadState(CIDRWHead * head);
  ~CIDRWHeadState() {}

  virtual int send(trigger * trig)
  {
    return this->send_trigger<ViewedState>(trig->name());
  }

  const cstr_id * get_current_name()
  {
    return this->current_state_name<ViewedState>();
  }

  static void InitEventDictionary();
  static task_base::event_dictionary _event_dictionary;

  bool m_operating;
};

struct InitHead : fsm::state<InitHead, CIDRWHeadState,
  fsm::custom_reaction<EvEvalCondition> >,
  NamedState
{
  InitHead(my_context ctx) : NamedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    post_event(boost::intrusive_ptr< EvEvalCondition >( new EvEvalCondition()) );
  }

  fsm::result react( const EvEvalCondition & )
  {
    if (outermost_context().m_operating)
      return transit<HeadOperating>();
    else
      return transit<HeadNotOperating>();
  }
};


struct HeadOperating : fsm::simple_state< HeadOperating, CIDRWHeadState,
  fsm::no_reactions, HeadIdle >,
  NamedState
{
    HeadOperating() : NamedState() { TRACE_STATE_ENTER(); }
};

struct HeadIdle : fsm::simple_state< HeadIdle, HeadOperating,
  mpl::list< fsm::transition< EvRead, HeadBusy >,
    fsm::transition< EvWrite, HeadBusy >,
    fsm::transition< EvDiagnostic, HeadBusy > > >,
  ViewedState
{
  HeadIdle() : ViewedState() { TRACE_STATE_ENTER(); }
};

struct HeadBusy : fsm::simple_state< HeadBusy, HeadOperating,
  fsm::transition< EvComplete, HeadIdle > >,
  ViewedState
{
  HeadBusy() : ViewedState() { TRACE_STATE_ENTER(); }
};

struct HeadNotOperating : fsm::simple_state< HeadNotOperating, CIDRWHeadState,
  fsm::transition< EvDiagnostic, HeadOperating > >,
  ViewedState
{
  HeadNotOperating() : ViewedState() {}
};


} }  // *namespace* jyugem::cidrw

#endif // JYUGEM_CIDRW_CIDRWSTATE_H

